/*
 * Copyright 2005-2007 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.solutions.identity.admin;

import org.wso2.solutions.identity.IdentityProviderException;
import org.wso2.utils.ServerConfiguration;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.security.KeyStore;
import java.security.cert.X509Certificate;

public class KeystoreUtilAdmin {

    public boolean isCertExists(X509Certificate cert) throws IdentityProviderException {

        ServerConfiguration serverConfig = ServerConfiguration.getInstance();
        String storeFilePath = serverConfig
                .getFirstProperty("Security.KeyStore.Location");
        String type = serverConfig.getFirstProperty("Security.KeyStore.Type");
        String passwd = serverConfig
                .getFirstProperty("Security.KeyStore.Password");

        try {
            FileInputStream is = new FileInputStream(storeFilePath);

            KeyStore store = KeyStore.getInstance(type);

            store.load(is, passwd.toCharArray());
            return store.containsAlias(getAlias(cert));
        } catch (Exception e) {
            throw new IdentityProviderException("errorInKeystoreAccess",
                    new String[] { storeFilePath }, e);
        }
    }
    
    public String importCert(X509Certificate cert)
            throws IdentityProviderException {
        ServerConfiguration serverConfig = ServerConfiguration.getInstance();
        String storeFilePath = serverConfig
                .getFirstProperty("Security.KeyStore.Location");
        String type = serverConfig.getFirstProperty("Security.KeyStore.Type");
        String passwd = serverConfig
                .getFirstProperty("Security.KeyStore.Password");
        return importCert(cert, storeFilePath, passwd, type);

    }

    public String importCert(X509Certificate cert, String storeFilePath,
            String passwd, String storeType) throws IdentityProviderException {
        try {

            FileInputStream is = new FileInputStream(storeFilePath);

            KeyStore store = KeyStore.getInstance(storeType);

            store.load(is, passwd.toCharArray());

            String alias = getAlias(cert);
            
            if(store.getCertificate(alias) != null) {
                throw new IdentityProviderException("relyingPartyExists",
                        new String[] { alias });
            }
            
            store.setCertificateEntry(alias, cert);

            FileOutputStream os = new FileOutputStream(storeFilePath);
            store.store(os, passwd.toCharArray());
            os.flush();

            return alias;
        } catch (Exception e) {
            throw new IdentityProviderException("errorInCertImport", e);
        }
    }

    public void removeCertEntry(String alias) throws IdentityProviderException {
        ServerConfiguration serverConfig = ServerConfiguration.getInstance();
        String storeFilePath = serverConfig
                .getFirstProperty("Security.KeyStore.Location");
        String type = serverConfig.getFirstProperty("Security.KeyStore.Type");
        String passwd = serverConfig
                .getFirstProperty("Security.KeyStore.Password");

        try {
            FileInputStream is = new FileInputStream(storeFilePath);

            KeyStore store = KeyStore.getInstance(type);

            store.load(is, passwd.toCharArray());
            
            if(store.containsAlias(alias) && !store.isKeyEntry(alias)) {
                //If this is not the private key entry remove it
                store.deleteEntry(alias);
            }
        } catch (Exception e) {
            throw new IdentityProviderException("errorInKeystoreAccess",
                    new String[] { storeFilePath }, e);
        }
    }
    
    private String getAlias(X509Certificate cert) throws Exception {
        // Alias should be the host name
        String name = cert.getSubjectDN().getName();
        String[] parts = name.split(",");
        String alias = null;
        for (int i = 0; i < parts.length; i++) {
            String cnStr = parts[i].trim();
            if (cnStr.startsWith("CN")) {
                alias = cnStr.substring(3);
                break;
            }
        }
        if (alias == null) {
            throw new IdentityProviderException("certNameInvalid",
                    new String[] { name });
        }
        return alias;
    }
    
}
